"""
Copyright 2021 AI Singapore

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

     https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import logging
from typing import Dict, Any, Tuple
import numpy as np

from peekingduck.weights_utils import checker, downloader
from peekingduck.pipeline.nodes.model.posenetv1.posenet_files.predictor import Predictor


class PoseNetModel:  # pylint: disable=too-few-public-methods
    """PoseNet model with model types: mobilenet50, mobilenet75, mobilenet101 and resnet"""

    def __init__(self, config: Dict[str, Any]) -> None:
        super().__init__()

        self.logger = logging.getLogger(__name__)

        # check for posenet weights, if none then download into weights folder
        if not checker.has_weights(config['root'],
                                   config['weights_dir']):
            print('---no posenet weights detected. proceeding to download...---')
            downloader.download_weights(config['root'],
                                        config['blob_file'])
            print('---posenet weights download complete.---')

        self.predictor = Predictor(config)

    def predict(self, frame: np.ndarray) -> \
            Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
        """ Predict poses from input frame

        Args:
            frame (np.array): image in numpy array

        Returns:
            bboxes, keypoints, keypoint_scores, keypoint_masks, keypoint_conns
            (Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]): \
            tuple containing list of bboxes and pose related info i.e coordinates,
            scores, connections
        """
        assert isinstance(frame, np.ndarray)

        return self.predictor.predict(frame)



"""
Copyright 2021 AI Singapore

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

     https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from typing import Any, Dict
from peekingduck.pipeline.nodes.node import AbstractNode
from peekingduck.pipeline.nodes.draw.utils.drawfunctions import draw_zones


class Node(AbstractNode):
    """Draw node for drawing zones onto image"""

    def __init__(self, config: Dict[str, Any]) -> None:
        super().__init__(config, node_path=__name__)

    def run(self, inputs: Dict[str, Any]) -> Dict[str, Any]:
        """Draws the boundaries of each specified zone onto the image.

        Args:
            inputs (dict): Dict with keys "zones", "img".

        Returns:
            outputs (dict): Dict with keys "img".
        """

        draw_zones(inputs["img"], inputs["zones"])  # type: ignore
        return {}



"""
Copyright 2018 Ross Wightman
Modifications copyright 2021 AI Singapore
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at
     https://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
from typing import Tuple
import numpy as np

from peekingduck.pipeline.nodes.model.posenetv1.posenet_files.constants import POSE_CONNECTIONS


def decode_pose(root_score: float,
                root_id: int,
                root_image_coord: np.ndarray,
                scores: np.ndarray,
                offsets: np.ndarray,
                output_stride: int,
                displacements_fwd: np.ndarray,
                displacements_bwd: np.ndarray,
                keypoint_scores: np.ndarray,
                keypoint_coords: np.ndarray) -> None:
    # pylint: disable=too-many-arguments
    """ Decode pose's keypoints scores and coordinates from keypoints score,
    coordinates and displacements
    Args:
        root_score (float): a keypoint with highest score is selected as root
        root_id (int): root keypoint's index
        root_image_coord (np.array): relative coordinate of root keypoint
        scores (np.array): HxWxNP heatmap scores of NP body parts
        offsets (np.array): HxWxNPx2 short range offset vector of NP body parts
        output_stride (int): output stride to convert output indices to image coordinates
        displacements_fwd (np.array): HxWxNEx2 forward displacements of NE body
                connections
        displacements_bwd (np.array): HxWxNEx2 backward displacements of NE body
                connections
        keypoints_scores (np.array): 17x1 buffer to store keypoint scores
        keypoint_coords (np.array): 17x2 buffer to store keypoint coordinates
    Returns:
        pose_count (int): number of poses detected
    """
    num_edges = len(POSE_CONNECTIONS)

    keypoint_scores[root_id] = root_score
    keypoint_coords[root_id] = root_image_coord

    for edge in reversed(range(num_edges)):
        target_keypoint_id, source_keypoint_id = POSE_CONNECTIONS[edge]
        _calculate_instance_keypoints(edge,
                                      target_keypoint_id,
                                      source_keypoint_id,
                                      keypoint_scores,
                                      keypoint_coords,
                                      scores,
                                      offsets,
                                      output_stride,
                                      displacements_bwd)

    for edge in range(num_edges):
        source_keypoint_id, target_keypoint_id = POSE_CONNECTIONS[edge]
        _calculate_instance_keypoints(edge,
                                      target_keypoint_id,
                                      source_keypoint_id,
                                      keypoint_scores,
                                      keypoint_coords,
                                      scores,
                                      offsets,
                                      output_stride,
                                      displacements_fwd)


def _calculate_instance_keypoints(edge: int,
                                  target_keypoint_id: int,
                                  source_keypoint_id: int,
                                  instance_keypoint_scores: np.ndarray,
                                  instance_keypoint_coords: np.ndarray,
                                  scores: np.ndarray,
                                  offsets: np.ndarray,
                                  output_stride: int,
                                  displacements: np.ndarray) -> None:
    # pylint: disable=too-many-arguments
    """ Obtain instance keypoints scores and coordinates
    """
    if (instance_keypoint_scores[source_keypoint_id] > 0.0 and
            instance_keypoint_scores[target_keypoint_id] == 0.0):
        source_keypoint = instance_keypoint_coords[source_keypoint_id]

        score, coords = _traverse_to_target_keypoint(edge,
                                                     source_keypoint,
                                                     target_keypoint_id,
                                                     scores,
                                                     offsets,
                                                     output_stride,
                                                     displacements)

        instance_keypoint_scores[target_keypoint_id] = score
        instance_keypoint_coords[target_keypoint_id] = coords


def _clip_to_indices(keypoints: np.ndarray,
                     output_stride: int,
                     width: int,
                     height: int) -> np.ndarray:
    """ Clip keypoint coordinate to indices within dimension (width, height)
    """
    keypoints = keypoints / output_stride
    keypoint_indices = np.zeros((2,), dtype=np.int32)

    keypoint_indices[0] = max(min(round(keypoints[0]), width - 1), 0)
    keypoint_indices[1] = max(min(round(keypoints[1]), height - 1), 0)

    return keypoint_indices


def _traverse_to_target_keypoint(edge_id: int,
                                 source_keypoint: np.ndarray,
                                 target_keypoint_id: int,
                                 scores: np.ndarray,
                                 offsets: np.ndarray,
                                 output_stride: int,
                                 displacements: np.ndarray) -> Tuple[float, float]:
    # pylint: disable=too-many-arguments
    """ Traverse to target keypoint to obtain keypoint score and coordinates
    """
    height = scores.shape[0] - 1
    width = scores.shape[1] - 1

    source_keypoint_indices = _clip_to_indices(
        source_keypoint, output_stride, width, height)

    displaced_point = source_keypoint + displacements[
        source_keypoint_indices[1], source_keypoint_indices[0], edge_id]

    displaced_point_indices = _clip_to_indices(
        displaced_point, output_stride, width, height)

    score = scores[displaced_point_indices[1],
                   displaced_point_indices[0], target_keypoint_id]

    image_coord = displaced_point_indices * output_stride + offsets[
        displaced_point_indices[1], displaced_point_indices[0], target_keypoint_id]

    return score, image_coord



"""
Copyright 2018 Ross Wightman
Modifications copyright 2021 AI Singapore
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at
     https://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
from typing import List
import numpy as np
import tensorflow as tf

from peekingduck.pipeline.nodes.model.posenetv1.posenet_files.decode_multi import \
    decode_multiple_poses


def get_keypoints_relative_coords(keypoint_coords: np.ndarray,
                                  output_scale: np.ndarray,
                                  image_size: List[int]) -> np.ndarray:
    """ Get keypoints coordinates relative to image size
    Args:
        keypoints_coords (np.array): Nx17x2 keypoints coordinates of N persons
        output_scale (np.array): output scale in Hx2 format
        image_size (List[int]): image size in HxW format
    Returns:
        keypoints_coords (np.array): Nx17x2 keypoints coordinates of N persons
                relative to image size
    """
    assert len(keypoint_coords.shape) == 3, "keypoint_coords should be 3D"
    assert keypoint_coords.shape[
        2], "keypoint_coords should be a 2D matrix of 2D offsets"
    rel_keypoint_coords = keypoint_coords * output_scale
    rel_keypoint_coords = rel_keypoint_coords / image_size
    return rel_keypoint_coords


def _sigmoid(array: np.ndarray) -> np.ndarray:
    return 1/(1 + np.exp(-array))


def detect_keypoints(
        tf_model: tf.keras.Model,
        image: tf.Tensor,
        output_stride: int,
        dst_scores: np.ndarray,
        dst_keypoints: np.ndarray,
        model_type: str,
        score_threshold: float) -> int:
    # pylint: disable=too-many-arguments
    """ Evaluate image by model function to get detected keypoints
    Args:
        tf_model: tensorflow model
        image (tf.Tensor): image for inference
        output_stride (int): output stride to convert output indices to image coordinates
        dst_scores (np.array): Nx17 buffer to store keypoint scores where N is
            the max persons to be detected
        dst_keypoints (np.array): Nx17x2 buffer to store keypoints coordinate
            where N is the max persons to be detected
        model_type (str): specified model type (refer to modelconfig.yml)
        score_threshold (float): threshold for prediction
    Returns:
        pose_count (int): number of poses detected
    """
    model_output = tf_model(image)

    # For resnet's implementation, we need to apply a sigmoid function on
    # the heatmap, which is the first tensor in the output
    if model_type == 'resnet':
        model_output[0] = _sigmoid(model_output[0])

    pose_count = decode_multiple_poses(
        model_output,
        dst_scores,
        dst_keypoints,
        output_stride=output_stride,
        min_pose_score=score_threshold)

    return pose_count



"""
Copyright 2021 AI Singapore

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

     https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Code of this file is mostly forked from
[@xuannianz](https://github.com/xuannianz))
"""

from typing import Dict, Any, Tuple

_KERAS_BACKEND = None
_KERAS_LAYERS = None
_KERAS_MODELS = None
_KERAS_UTILS = None


def get_submodules_from_kwargs(kwargs: Dict[str, Any]) -> Tuple[Any, Any, Any, Any]:
    """Helper function to get keras submodules

    Args:
        kwargs (dict]): dictionary of keyword args

    Returns:
        (tuple): the specified submodules
    """
    backend = kwargs.get('backend', _KERAS_BACKEND)
    layers = kwargs.get('layers', _KERAS_LAYERS)
    models = kwargs.get('models', _KERAS_MODELS)
    utils = kwargs.get('utils', _KERAS_UTILS)
    for key in kwargs.keys():
        if key not in ['backend', 'layers', 'models', 'utils']:
            raise TypeError(f'Invalid keyword argument: {key}')
    return backend, layers, models, utils



"""
Copyright 2021 AI Singapore

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

     https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
from typing import Dict, Any
from peekingduck.pipeline.nodes.node import AbstractNode
from peekingduck.pipeline.nodes.model.efficientdet_d04 import efficientdet_model


class Node(AbstractNode):
    """EfficientDet node class that initializes and uses efficientdet model to detect
    bounding boxes from an image.

    EfficientDet: Scalable and Efficient Object Detection
    https://arxiv.org/abs/1911.09070
    """

    def __init__(self, config: Dict[str, Any]) -> None:
        super().__init__(config, node_path=__name__)
        self.model = efficientdet_model.EfficientDetModel(config)

    def run(self, inputs: Dict[str, Any]) -> Dict[str, Any]:
        """Function that takes an image as input and returns bboxes of objects specified
        in config.

        Args:
            inputs (Dict): Dict with key "img"

        Returns:
            outputs (Dict): Dict with keys "bboxes".
        """
        # Currently prototyped to return just the bounding boxes
        # without the scores
        bboxes, labels, scores = self.model.predict(inputs["img"])
        outputs = {"bboxes": bboxes, "bbox_labels": labels, "bbox_scores": scores}
        return outputs



"""
Copyright 2021 AI Singapore

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

     https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Code of this file is mostly forked from
[@xuannianz](https://github.com/xuannianz))
"""

from typing import List, Tuple
import numpy as np
import cv2

IMG_MEAN = [0.485, 0.456, 0.406]
IMG_STD = [0.229, 0.224, 0.225]


def preprocess_image(image: np.ndarray,
                     image_size: int) -> Tuple[List[List[float]], float]:
    """Preprocessing helper function for efficientdet

    Args:
        image (np.array): the input image in numpy array
        image_size (int): the model input size as specified in config

    Returns:
        image (np.array): the preprocessed image
        scale (float): the scale in which the original image was resized to
    """
    # image, RGB
    image_height, image_width = image.shape[:2]
    if image_height > image_width:
        scale = image_size / image_height
        resized_height = image_size
        resized_width = int(image_width * scale)
    else:
        scale = image_size / image_width
        resized_height = int(image_height * scale)
        resized_width = image_size

    image = cv2.resize(image, (resized_width, resized_height))
    image = image.astype(np.float32)
    image /= 255.
    image -= IMG_MEAN
    image /= IMG_STD
    pad_h = image_size - resized_height
    pad_w = image_size - resized_width
    image = np.pad(image, [(0, pad_h), (0, pad_w), (0, 0)], mode='constant')

    return image, scale


def postprocess_boxes(boxes: np.ndarray,
                      scale: float,
                      height: int,
                      width: int) -> np.ndarray:
    """Postprocessing helper function for efficientdet

    Args:
        boxes (np.array): the original detected bboxes from model output
        scale (float): scale in which the original image was resized to
        height (int): the height of the original image
        width (int): the width of the original image

    Returns:
        boxes (np.array): the postprocessed bboxes
    """
    boxes /= scale
    boxes[:, 0] = np.clip(boxes[:, 0], 0, width - 1)
    boxes[:, 1] = np.clip(boxes[:, 1], 0, height - 1)
    boxes[:, 2] = np.clip(boxes[:, 2], 0, width - 1)
    boxes[:, 3] = np.clip(boxes[:, 3], 0, height - 1)

    boxes[:, [0, 2]] /= width
    boxes[:, [1, 3]] /= height
    return boxes



"""
Copyright 2019 The TensorFlow Authors, Pavel Yakubovskiy, Björn Barz. All Rights Reserved.

Modifications copyright 2021 AI Singapore

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
==============================================================================
Contains definitions for EfficientNet model.

[1] Mingxing Tan, Quoc V. Le
  EfficientNet: Rethinking Model Scaling for Convolutional Neural Networks.
  ICML'19, https://arxiv.org/abs/1905.11946

Code of this model implementation is mostly written by
Björn Barz ([@Callidior](https://github.com/Callidior))
"""


from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os
import math
import string
import collections
from typing import Any, Dict, Callable, List, Tuple, Union

from six.moves import xrange
import tensorflow as tf
from tensorflow.python.keras.applications.imagenet_utils import obtain_input_shape
from tensorflow.keras.applications.imagenet_utils import preprocess_input as _preprocess_input

from peekingduck.pipeline.nodes.model.efficientdet_d04.efficientdet_files.utils.submodule \
    import get_submodules_from_kwargs


BACKEND = None
LAYERS = None
MODELS = None
KERAS_UTILS = None


BASE_WEIGHTS_PATH = (
    'https://github.com/Callidior/keras-applications/'
    'releases/download/efficientnet/')

WEIGHTS_HASHES = {
    'efficientnet-b0': ('163292582f1c6eaca8e7dc7b51b01c61'
                        '5b0dbc0039699b4dcd0b975cc21533dc',
                        'c1421ad80a9fc67c2cc4000f666aa507'
                        '89ce39eedb4e06d531b0c593890ccff3'),
    'efficientnet-b1': ('d0a71ddf51ef7a0ca425bab32b7fa7f1'
                        '6043ee598ecee73fc674d9560c8f09b0',
                        '75de265d03ac52fa74f2f510455ba64f'
                        '9c7c5fd96dc923cd4bfefa3d680c4b68'),
    'efficientnet-b2': ('bb5451507a6418a574534aa76a91b106'
                        'f6b605f3b5dde0b21055694319853086',
                        '433b60584fafba1ea3de07443b74cfd3'
                        '2ce004a012020b07ef69e22ba8669333'),
    'efficientnet-b3': ('03f1fba367f070bd2545f081cfa7f3e7'
                        '6f5e1aa3b6f4db700f00552901e75ab9',
                        'c5d42eb6cfae8567b418ad3845cfd63a'
                        'a48b87f1bd5df8658a49375a9f3135c7'),
    'efficientnet-b4': ('98852de93f74d9833c8640474b2c698d'
                        'b45ec60690c75b3bacb1845e907bf94f',
                        '7942c1407ff1feb34113995864970cd4'
                        'd9d91ea64877e8d9c38b6c1e0767c411'),
    'efficientnet-b5': ('30172f1d45f9b8a41352d4219bf930ee'
                        '3339025fd26ab314a817ba8918fefc7d',
                        '9d197bc2bfe29165c10a2af8c2ebc675'
                        '07f5d70456f09e584c71b822941b1952'),
    'efficientnet-b6': ('f5270466747753485a082092ac9939ca'
                        'a546eb3f09edca6d6fff842cad938720',
                        '1d0923bb038f2f8060faaf0a0449db4b'
                        '96549a881747b7c7678724ac79f427ed'),
    'efficientnet-b7': ('876a41319980638fa597acbbf956a82d'
                        '10819531ff2dcb1a52277f10c7aefa1a',
                        '60b56ff3a8daccc8d96edfd40b204c11'
                        '3e51748da657afd58034d54d3cec2bac')
}

BlockArgs = collections.namedtuple('BlockArgs', [
    'kernel_size', 'num_repeat', 'input_filters', 'output_filters',
    'expand_ratio', 'id_skip', 'strides', 'se_ratio'
])
# defaults will be a public argument for namedtuple in Python 3.7
# https://docs.python.org/3/library/collections.html#collections.namedtuple
# mypy: ignore-errors
BlockArgs.__new__.__defaults__ = (None,) * len(BlockArgs._fields)  # type: ignore

DEFAULT_BLOCKS_ARGS = (
    BlockArgs(kernel_size=3, num_repeat=1, input_filters=32, output_filters=16,
              expand_ratio=1, id_skip=True, strides=[1, 1], se_ratio=0.25),
    BlockArgs(kernel_size=3, num_repeat=2, input_filters=16, output_filters=24,
              expand_ratio=6, id_skip=True, strides=[2, 2], se_ratio=0.25),
    BlockArgs(kernel_size=5, num_repeat=2, input_filters=24, output_filters=40,
              expand_ratio=6, id_skip=True, strides=[2, 2], se_ratio=0.25),
    BlockArgs(kernel_size=3, num_repeat=3, input_filters=40, output_filters=80,
              expand_ratio=6, id_skip=True, strides=[2, 2], se_ratio=0.25),
    BlockArgs(kernel_size=5, num_repeat=3, input_filters=80, output_filters=112,
              expand_ratio=6, id_skip=True, strides=[1, 1], se_ratio=0.25),
    BlockArgs(kernel_size=5, num_repeat=4, input_filters=112, output_filters=192,
              expand_ratio=6, id_skip=True, strides=[2, 2], se_ratio=0.25),
    BlockArgs(kernel_size=3, num_repeat=1, input_filters=192, output_filters=320,
              expand_ratio=6, id_skip=True, strides=[1, 1], se_ratio=0.25)
)

CONV_KERNEL_INITIALIZER = {
    'class_name': 'VarianceScaling',
    'config': {
        'scale': 2.0,
        'mode': 'fan_out',
        # EfficientNet actually uses an untruncated normal distribution for
        # initializing conv layers, but keras.initializers.VarianceScaling use
        # a truncated distribution.
        # We decided against a custom initializer for better serializability.
        'distribution': 'normal'
    }
}

DENSE_KERNEL_INITIALIZER = {
    'class_name': 'VarianceScaling',
    'config': {
        'scale': 1. / 3.,
        'mode': 'fan_out',
        'distribution': 'uniform'
    }
}


def preprocess_input(model_input: tf.Tensor, **kwargs: Dict[str, Any]) -> tf.Tensor:
    """Preprocesses model input
    """
    kwargs = {k: v for k, v in kwargs.items() if k in ['backend', 'layers', 'models', 'utils']}
    return _preprocess_input(model_input, mode='torch', **kwargs)


def get_swish(**kwargs: Dict[str, Any]) -> Callable:
    """Swish activation function: x * sigmoid(x).
        Reference: [Searching for Activation Functions](https://arxiv.org/abs/1710.05941)
        """
    backend, _, _, _ = get_submodules_from_kwargs(kwargs)

    def swish(swish_x: tf.Tensor) -> tf.Tensor:

        if backend.backend() == 'tensorflow':
            try:
                # The native TF implementation has a more
                # memory-efficient gradient implementation
                return backend.tf.nn.swish(swish_x)
            except AttributeError:
                pass

        return swish_x * backend.sigmoid(swish_x)

    return swish


def get_dropout(**kwargs: Union[None, Dict[str, Any]]) -> Any:
    """Wrapper over custom dropout. Fix problem of ``None`` shape for tf.keras.
    It is not possible to define FixedDropout class as global object,
    because we do not have modules for inheritance at first time.

    Issue:
        https://github.com/tensorflow/tensorflow/issues/30946
    """
    backend, layers, _, _ = get_submodules_from_kwargs(kwargs)

    class FixedDropout(layers.Dropout):  # pylint: disable=too-few-public-methods
        """Fixed Dropout Class
        """

        def _get_noise_shape(self, inputs: tf.Tensor) -> Union[None, Tuple[Any, ...]]:
            if self.noise_shape is None:
                return self.noise_shape

            symbolic_shape = backend.shape(inputs)
            noise_shape = [symbolic_shape[axis] if shape is None else shape
                           for axis, shape in enumerate(self.noise_shape)]
            return tuple(noise_shape)

    return FixedDropout


def round_filters(filters: float, width_coefficient: float, depth_divisor: int) -> int:
    """Round number of filters based on width multiplier."""

    filters *= width_coefficient
    new_filters = int(filters + depth_divisor / 2) // depth_divisor * depth_divisor
    new_filters = max(depth_divisor, new_filters)
    # Make sure that round down does not go down by more than 10%.
    if new_filters < 0.9 * filters:
        new_filters += depth_divisor
    return int(new_filters)


def round_repeats(repeats: int, depth_coefficient: float) -> int:
    """Round number of repeats based on depth multiplier."""

    return int(math.ceil(depth_coefficient * repeats))


def mb_conv_block(inputs: tf.Tensor,  # type:ignore
                  block_args,
                  activation: Callable,
                  drop_rate: float = None,
                  prefix: str = '') -> tf.Tensor:
    """Mobile Inverted Residual Bottleneck."""

    has_se = (block_args.se_ratio is not None) and (0 < block_args.se_ratio <= 1)
    bn_axis = 3 if BACKEND.image_data_format() == 'channels_last' else 1

    # workaround over non working dropout with None in noise_shape in tf.keras
    Dropout = get_dropout(
        backend=BACKEND,
        layers=LAYERS,
        models=MODELS,
        utils=KERAS_UTILS
    )

    # Expansion phase
    filters = block_args.input_filters * block_args.expand_ratio
    if block_args.expand_ratio != 1:
        x_in = LAYERS.Conv2D(filters, 1,
                             padding='same',
                             use_bias=False,
                             kernel_initializer=CONV_KERNEL_INITIALIZER,
                             name=prefix + 'expand_conv')(inputs)
        x_in = LAYERS.BatchNormalization(axis=bn_axis, name=prefix + 'expand_bn')(x_in)
        x_in = LAYERS.Activation(activation, name=prefix + 'expand_activation')(x_in)
    else:
        x_in = inputs

    # Depthwise Convolution
    x_in = LAYERS.DepthwiseConv2D(block_args.kernel_size,
                                  strides=block_args.strides,
                                  padding='same',
                                  use_bias=False,
                                  depthwise_initializer=CONV_KERNEL_INITIALIZER,
                                  name=prefix + 'dwconv')(x_in)
    x_in = LAYERS.BatchNormalization(axis=bn_axis, name=prefix + 'bn')(x_in)
    x_in = LAYERS.Activation(activation, name=prefix + 'activation')(x_in)

    # Squeeze and Excitation phase
    if has_se:
        num_reduced_filters = max(1, int(
            block_args.input_filters * block_args.se_ratio
        ))
        se_tensor = LAYERS.GlobalAveragePooling2D(name=prefix + 'se_squeeze')(x_in)

        target_shape = (1, 1, filters) if BACKEND.image_data_format(
        ) == 'channels_last' else (filters, 1, 1)
        se_tensor = LAYERS.Reshape(target_shape, name=prefix + 'se_reshape')(se_tensor)
        se_tensor = LAYERS.Conv2D(num_reduced_filters, 1,
                                  activation=activation,
                                  padding='same',
                                  use_bias=True,
                                  kernel_initializer=CONV_KERNEL_INITIALIZER,
                                  name=prefix + 'se_reduce')(se_tensor)
        se_tensor = LAYERS.Conv2D(filters, 1,
                                  activation='sigmoid',
                                  padding='same',
                                  use_bias=True,
                                  kernel_initializer=CONV_KERNEL_INITIALIZER,
                                  name=prefix + 'se_expand')(se_tensor)
        if BACKEND.backend() == 'theano':
            # For the Theano backend, we have to explicitly make
            # the excitation weights broadcastable.
            pattern = ([True, True, True, False] if BACKEND.image_data_format() == 'channels_last'
                       else [True, False, True, True])
            se_tensor = LAYERS.Lambda(
                lambda x: BACKEND.pattern_broadcast(x, pattern),
                name=prefix + 'se_broadcast')(se_tensor)
        x_in = LAYERS.multiply([x_in, se_tensor], name=prefix + 'se_excite')

    # Output phase
    x_in = LAYERS.Conv2D(block_args.output_filters, 1,
                         padding='same',
                         use_bias=False,
                         kernel_initializer=CONV_KERNEL_INITIALIZER,
                         name=prefix + 'project_conv')(x_in)
    x_in = LAYERS.BatchNormalization(axis=bn_axis, name=prefix + 'project_bn')(x_in)
    if block_args.id_skip and all(
            s == 1 for s in block_args.strides
    ) and block_args.input_filters == block_args.output_filters:
        if drop_rate and (drop_rate > 0):
            x_in = Dropout(drop_rate,
                           noise_shape=(None, 1, 1, 1),
                           name=prefix + 'drop')(x_in)
        x_in = LAYERS.add([x_in, inputs], name=prefix + 'add')

    return x_in


def efficientnet_base(width_coefficient: float,  # pylint: disable=too-many-arguments, too-many-locals, too-many-branches
                      depth_coefficient: float,
                      default_resolution: int,
                      drop_connect_rate: float = 0.2,
                      depth_divisor: int = 8,
                      blocks_args=DEFAULT_BLOCKS_ARGS,
                      include_top: bool = True,
                      weights: str = 'imagenet',
                      input_tensor: Union[None, tf.Tensor] = None,
                      input_shape: Union[None, Tuple[int, int, int]] = None,
                      classes: int = 1000,
                      **kwargs: Any) -> List[tf.Tensor]:
    """Instantiates the EfficientNet architecture using given scaling coefficients.
    Optionally loads weights pre-trained on ImageNet.
    Note that the data format convention used by the model is
    the one specified in your Keras config at `~/.keras/keras.json`.
    # Arguments
        width_coefficient: float, scaling coefficient for network width.
        depth_coefficient: float, scaling coefficient for network depth.
        default_resolution: int, default input image size.
        dropout_rate: float, dropout rate before final classifier layer.
        drop_connect_rate: float, dropout rate at skip connections.
        depth_divisor: int.
        blocks_args: A tuple of BlockArgs to construct block modules.
        model_name: string, model name.
        include_top: whether to include the fully-connected
            layer at the top of the network.
        weights: one of `None` (random initialization),
              'imagenet' (pre-training on ImageNet),
              or the path to the weights file to be loaded.
        input_tensor: optional Keras tensor
            (i.e. output of `layers.Input()`)
            to use as image input for the model.
        input_shape: optional shape tuple, only to be specified
            if `include_top` is False.
            It should have exactly 3 inputs channels.
        pooling: optional pooling mode for feature extraction
            when `include_top` is `False`.
            - `None` means that the output of the model will be
                the 4D tensor output of the
                last convolutional layer.
            - `avg` means that global average pooling
                will be applied to the output of the
                last convolutional layer, and thus
                the output of the model will be a 2D tensor.
            - `max` means that global max pooling will
                be applied.
        classes: optional number of classes to classify images
            into, only to be specified if `include_top` is True, and
            if no `weights` argument is specified.
    # Returns
        A Keras model instance.
    # Raises
        ValueError: in case of invalid argument for `weights`,
            or invalid input shape.
    """
    global BACKEND, LAYERS, MODELS, KERAS_UTILS  # pylint: disable=global-statement
    BACKEND, LAYERS, MODELS, KERAS_UTILS = get_submodules_from_kwargs(kwargs)
    features = []
    if not (weights in {'imagenet', None} or os.path.exists(weights)):
        raise ValueError('The `weights` argument should be either '
                         '`None` (random initialization), `imagenet` '
                         '(pre-training on ImageNet), '
                         'or the path to the weights file to be loaded.')

    if weights == 'imagenet' and include_top and classes != 1000:
        raise ValueError('If using `weights` as `"imagenet"` with `include_top`'
                         ' as true, `classes` should be 1000')

    # Determine proper input shape
    input_shape = obtain_input_shape(input_shape,
                                     default_size=default_resolution,
                                     min_size=32,
                                     data_format=BACKEND.image_data_format(),
                                     require_flatten=include_top,
                                     weights=weights)

    if input_tensor is None:
        img_input = LAYERS.Input(shape=input_shape)
    else:
        if BACKEND.backend() == 'tensorflow':
            from tensorflow.python.keras.backend import is_keras_tensor  # pylint: disable=import-outside-toplevel
        else:
            is_keras_tensor = BACKEND.is_keras_tensor
        if not is_keras_tensor(input_tensor):
            img_input = LAYERS.Input(tensor=input_tensor, shape=input_shape)
        else:
            img_input = input_tensor

    bn_axis = 3 if BACKEND.image_data_format() == 'channels_last' else 1
    activation = get_swish(**kwargs)

    # Build stem
    x_in = img_input
    x_in = LAYERS.Conv2D(round_filters(32, width_coefficient, depth_divisor), 3,
                         strides=(2, 2),
                         padding='same',
                         use_bias=False,
                         kernel_initializer=CONV_KERNEL_INITIALIZER,
                         name='stem_conv')(x_in)
    x_in = LAYERS.BatchNormalization(axis=bn_axis, name='stem_bn')(x_in)
    x_in = LAYERS.Activation(activation, name='stem_activation')(x_in)
    # Build blocks
    num_blocks_total = sum(block_args.num_repeat for block_args in blocks_args)
    block_num = 0
    for idx, block_args in enumerate(blocks_args):
        assert block_args.num_repeat > 0
        # Update block input and output filters based on depth multiplier.
        block_args = block_args._replace(
            input_filters=round_filters(block_args.input_filters,
                                        width_coefficient, depth_divisor),
            output_filters=round_filters(block_args.output_filters,
                                         width_coefficient, depth_divisor),
            num_repeat=round_repeats(block_args.num_repeat, depth_coefficient))

        # The first block needs to take care of stride and filter size increase.
        drop_rate = drop_connect_rate * float(block_num) / num_blocks_total
        x_in = mb_conv_block(x_in, block_args,
                             activation=activation,
                             drop_rate=drop_rate,
                             prefix='block{}a_'.format(idx + 1))
        block_num += 1
        if block_args.num_repeat > 1:
            # pylint: disable=protected-access
            block_args = block_args._replace(
                input_filters=block_args.output_filters, strides=[1, 1])
            # pylint: enable=protected-access
            for bidx in xrange(block_args.num_repeat - 1):
                drop_rate = drop_connect_rate * float(block_num) / num_blocks_total
                block_prefix = 'block{}{}_'.format(
                    idx + 1,
                    string.ascii_lowercase[bidx + 1]
                )
                x_in = mb_conv_block(x_in, block_args,
                                     activation=activation,
                                     drop_rate=drop_rate,
                                     prefix=block_prefix)
                block_num += 1
        if idx < len(blocks_args) - 1 and blocks_args[idx + 1].strides[0] == 2:
            features.append(x_in)
        elif idx == len(blocks_args) - 1:
            features.append(x_in)
    return features


def efficientnet_b0(include_top: bool = True,
                    weights: str = 'imagenet',
                    input_tensor: Union[None, tf.Tensor] = None,
                    input_shape: Union[None, Tuple[int, int, int]] = None,
                    classes: int = 1000,
                    **kwargs: Any) -> List[tf.Tensor]:
    """EfficientNet-B0 model
    """
    return efficientnet_base(1.0, 1.0, 224, 0.2,
                             include_top=include_top, weights=weights,
                             input_tensor=input_tensor, input_shape=input_shape,
                             classes=classes,
                             **kwargs)


def efficientnet_b1(include_top: bool = True,
                    weights: str = 'imagenet',
                    input_tensor: Union[None, tf.Tensor] = None,
                    input_shape: Union[None, Tuple[int, int, int]] = None,
                    classes: int = 1000,
                    **kwargs: Any) -> List[tf.Tensor]:
    """EfficientNet-B1 model
    """
    return efficientnet_base(1.0, 1.1, 240, 0.2,
                             include_top=include_top, weights=weights,
                             input_tensor=input_tensor, input_shape=input_shape,
                             classes=classes, **kwargs)


def efficientnet_b2(include_top: bool = True,
                    weights: str = 'imagenet',
                    input_tensor: Union[None, tf.Tensor] = None,
                    input_shape: Union[None, Tuple[int, int, int]] = None,
                    classes: int = 1000,
                    **kwargs: Any) -> List[tf.Tensor]:
    """EfficientNet-B2 model
    """
    return efficientnet_base(1.1, 1.2, 260, 0.3,
                             include_top=include_top, weights=weights,
                             input_tensor=input_tensor, input_shape=input_shape,
                             classes=classes,
                             **kwargs)


def efficientnet_b3(include_top: bool = True,
                    weights: str = 'imagenet',
                    input_tensor: Union[None, tf.Tensor] = None,
                    input_shape: Union[None, Tuple[int, int, int]] = None,
                    classes: int = 1000,
                    **kwargs: Any) -> List[tf.Tensor]:
    """EfficientNet-B3 model
    """
    return efficientnet_base(1.2, 1.4, 300, 0.3,
                             include_top=include_top, weights=weights,
                             input_tensor=input_tensor, input_shape=input_shape,
                             classes=classes,
                             **kwargs)


def efficientnet_b4(include_top: bool = True,
                    weights: str = 'imagenet',
                    input_tensor: Union[None, tf.Tensor] = None,
                    input_shape: Union[None, Tuple[int, int, int]] = None,
                    classes: int = 1000,
                    **kwargs: Any) -> List[tf.Tensor]:
    """EfficientNet-B4 model
    """
    return efficientnet_base(1.4, 1.8, 380, 0.4,
                             include_top=include_top, weights=weights,
                             input_tensor=input_tensor, input_shape=input_shape,
                             classes=classes,
                             **kwargs)


def efficientnet_b5(include_top: bool = True,
                    weights: str = 'imagenet',
                    input_tensor: Union[None, tf.Tensor] = None,
                    input_shape: Union[None, Tuple[int, int, int]] = None,
                    classes: int = 1000,
                    **kwargs: Any) -> List[tf.Tensor]:
    """EfficientNet-B5 model
    """
    return efficientnet_base(1.6, 2.2, 456, 0.4,
                             include_top=include_top, weights=weights,
                             input_tensor=input_tensor, input_shape=input_shape,
                             classes=classes,
                             **kwargs)


def efficientnet_b6(include_top: bool = True,
                    weights: str = 'imagenet',
                    input_tensor: Union[None, tf.Tensor] = None,
                    input_shape: Union[None, Tuple[int, int, int]] = None,
                    classes: int = 1000,
                    **kwargs: Any) -> List[tf.Tensor]:
    """EfficientNet-B6 model
    """
    return efficientnet_base(1.8, 2.6, 528, 0.5,
                             include_top=include_top, weights=weights,
                             input_tensor=input_tensor, input_shape=input_shape,
                             classes=classes,
                             **kwargs)


def efficientnet_b7(include_top: bool = True,
                    weights: str = 'imagenet',
                    input_tensor: Union[None, tf.Tensor] = None,
                    input_shape: Union[None, Tuple[int, int, int]] = None,
                    classes: int = 1000,
                    **kwargs: Any) -> List[tf.Tensor]:
    """EfficientNet-B7 model
    """
    return efficientnet_base(2.0, 3.1, 600, 0.5,
                             include_top=include_top, weights=weights,
                             input_tensor=input_tensor, input_shape=input_shape,
                             classes=classes,
                             **kwargs)


setattr(efficientnet_b0, '__doc__', efficientnet_base.__doc__)
setattr(efficientnet_b1, '__doc__', efficientnet_base.__doc__)
setattr(efficientnet_b2, '__doc__', efficientnet_base.__doc__)
setattr(efficientnet_b3, '__doc__', efficientnet_base.__doc__)
setattr(efficientnet_b4, '__doc__', efficientnet_base.__doc__)
setattr(efficientnet_b5, '__doc__', efficientnet_base.__doc__)
setattr(efficientnet_b6, '__doc__', efficientnet_base.__doc__)
setattr(efficientnet_b7, '__doc__', efficientnet_base.__doc__)



"""
Copyright 2021 AI Singapore

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

     https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from typing import Dict, Any
import numpy as np
from peekingduck.pipeline.nodes.node import AbstractNode

NOSE = 0
LEFT_SHOULDER = 5
RIGHT_SHOULDER = 6
LEFT_PELVIS = 11
RIGHT_PELVIS = 12
TORSO_KEYPOINTS = [
    NOSE, LEFT_SHOULDER, RIGHT_SHOULDER, LEFT_PELVIS, RIGHT_PELVIS]


class Node(AbstractNode):
    """Node that uses pose keypoint information of torso to estimate 3d location"""

    def __init__(self, config: Dict[str, Any]) -> None:
        super().__init__(config, node_path=__name__)

        self.torso_factor = config['torso_factor']
        self.focal_length = config['focal_length']

    def run(self, inputs: Dict[str, Any]) -> Dict[str, Any]:
        """Converts pose keypoints into 3D locations.

        Args:
            inputs (dict): Dict with keys "keypoints".

        Returns:
            outputs (dict): Dict with keys "obj_3D_locs".
        """

        locations = []

        for keypoints in inputs["keypoints"]:
            torso_keypoints = self._get_torso_keypoints(keypoints)
            if self._enough_torso_keypoints(torso_keypoints):
                bbox = self._get_bbox(torso_keypoints)
            else:
                bbox = self._get_bbox(keypoints)

            point = self._get_3d_point_from_bbox(
                bbox, self.focal_length, self.torso_factor)
            locations.append(point)

        outputs = {"obj_3D_locs": locations}

        return outputs

    @staticmethod
    def _get_torso_keypoints(keypoints: np.array) -> np.array:
        """Filter keypoints to get only selected keypoints for torso"""

        torso_keypoints = keypoints[TORSO_KEYPOINTS, :]  # type: ignore
        # ignore keypoints that are '-1.' as below confidence score and are masked
        torso_keypoints = np.reshape(
            torso_keypoints[torso_keypoints != -1.], (-1, 2))

        return torso_keypoints

    @staticmethod
    def _enough_torso_keypoints(torso_keypoints: np.array) -> bool:
        """Returns False if not enough keypoints to represent torso"""

        if torso_keypoints.shape[0] >= 2:
            return True
        return False

    @staticmethod
    def _get_bbox(keypoints: np.array) -> np.array:
        """Get coordinates of a bbox around keypoints"""

        top_left_x, top_left_y = keypoints.min(axis=0)
        btm_right_x, btm_right_y = keypoints.max(axis=0)

        return np.array([top_left_x, top_left_y, btm_right_x, btm_right_y])

    @staticmethod
    def _get_3d_point_from_bbox(bbox: np.array, focal_length: float,
                                torso_factor: float) -> np.array:
        """Get the 3d coordinates of the centre of a bounding box"""

        # Subtraction is to make the camera the origin of the coordinate system
        center_2d = ((bbox[0:2] + bbox[2:4]) * 0.5) - np.array([0.5, 0.5])
        torso_height = bbox[3] - bbox[1]

        z_coord = (focal_length * torso_factor) / torso_height
        x_coord = (center_2d[0] * torso_factor) / torso_height
        y_coord = (center_2d[1] * torso_factor) / torso_height

        return np.array([x_coord, y_coord, z_coord])



"""
Copyright 2021 AI Singapore

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

     https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Code of this file is mostly forked from
[@xuannianz](https://github.com/xuannianz))
"""
from typing import Any, Dict, List
from tensorflow import keras
import tensorflow as tf


class WBiFPNAdd(keras.layers.Layer):
    """Class for Weighted Bi-directional FPN
    """

    def __init__(self, epsilon: float = 1e-4, **kwargs: Any) -> None:
        super().__init__(**kwargs)
        self.epsilon = epsilon
        self.weight = None

    def build(self, input_shape: List[tf.TensorShape]) -> None:
        num_in = len(input_shape)
        self.weight = self.add_weight(name=self.name,
                                      shape=(num_in,),
                                      initializer=keras.initializers.constant(1 / num_in),
                                      trainable=True,
                                      dtype=tf.float32)

    def call(self, inputs: List[tf.Tensor], **kwargs: Dict[str, Any]) -> tf.Tensor:
        weight = keras.activations.relu(self.weight)
        x_in = tf.reduce_sum([weight[i] * inputs[i] for i in range(len(inputs))], axis=0)
        x_in = x_in / (tf.reduce_sum(weight) + self.epsilon)
        return x_in

    def compute_output_shape(self, input_shape: List[tf.TensorShape]) -> tf.TensorShape:
        return input_shape[0]

    def get_config(self) -> Dict[str, Any]:
        config = super().get_config()
        config.update({
            'epsilon': self.epsilon
        })
        return config


def bbox_transform_inv(boxes: tf.Tensor,
                       deltas: tf.Tensor,
                       scale_factors: List[float] = None) -> tf.Tensor:
    """Helper function to transform bboxes using offsets

    Args:
        boxes : detected bboxes]
        deltas : bbox offsets
        scale_factors : List of scales for each dimension

    Returns:
        tf.Tensor: bboxes in xmin, ymin, xmax, ymax format
    """

    wt_a = boxes[..., 2] - boxes[..., 0]
    ht_a = boxes[..., 3] - boxes[..., 1]
    t_y, t_x, t_h, t_w = deltas[..., 0], deltas[..., 1], deltas[..., 2], deltas[..., 3]
    if scale_factors:
        t_y *= scale_factors[0]
        t_x *= scale_factors[1]
        t_h *= scale_factors[2]
        t_w *= scale_factors[3]
    width = tf.exp(t_w) * wt_a
    height = tf.exp(t_h) * ht_a
    center_y = t_y * ht_a + (boxes[..., 1] + boxes[..., 3]) / 2
    center_x = t_x * wt_a + (boxes[..., 0] + boxes[..., 2]) / 2
    top_left = center_x - width / 2., center_y - height / 2.
    btm_right = center_x + width / 2., center_y + height / 2.
    # return tf.stack([xmin, ymin, xmax, ymax], axis=-1)
    return tf.stack([top_left[0], top_left[1], btm_right[0], btm_right[1]], axis=-1)


class ClipBoxes(keras.layers.Layer):
    """ClipBoxes class to limit the value of bbox coordinates to image height and width
    """

    def call(self, inputs: List[tf.Tensor], **kwargs: Dict[str, Any]) -> tf.Tensor:
        image, boxes = inputs
        shape = keras.backend.cast(keras.backend.shape(image), keras.backend.floatx())
        height = shape[1]
        width = shape[2]
        x_1 = tf.clip_by_value(boxes[:, :, 0], 0, width - 1)
        y_1 = tf.clip_by_value(boxes[:, :, 1], 0, height - 1)
        x_2 = tf.clip_by_value(boxes[:, :, 2], 0, width - 1)
        y_2 = tf.clip_by_value(boxes[:, :, 3], 0, height - 1)

        return keras.backend.stack([x_1, y_1, x_2, y_2], axis=2)

    def compute_output_shape(self, input_shape: List[tf.TensorShape]) -> tf.TensorShape:
        return input_shape[1]


class RegressBoxes(keras.layers.Layer):
    """RegressBoxes class to compute actual bbox coordinate using anchors and offsets
    """

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

    def call(self, inputs: List[tf.Tensor], **kwargs: Dict[str, Any]) -> tf.Tensor:
        anchors, regression = inputs
        return bbox_transform_inv(anchors, regression)

    def compute_output_shape(self, input_shape: List[tf.TensorShape]) -> tf.TensorShape:
        return input_shape[0]

    def get_config(self) -> Dict[str, Any]:
        config = super().get_config()
        return config


def filter_detections(  # pylint: disable=too-many-arguments, too-many-locals
        boxes: tf.Tensor,
        classification: tf.Tensor,
        alphas: tf.Tensor = None,
        ratios: tf.Tensor = None,
        class_specific_filter: bool = True,
        nms: bool = True,
        score_threshold: float = 0.01,
        max_detections: int = 100,
        nms_threshold: float = 0.5,
        detect_quadrangle: bool = False,
) -> List[tf.Tensor]:
    """
    Filter detections using the boxes and classification values.

    Args
        boxes: Tensor of shape (num_boxes, 4) containing the boxes in (x1, y1, x2, y2) format.
        classification: Tensor of shape (num_boxes, num_classes) containing classification scores.
        other: List of tensors (num_boxes, ...) to filter the boxes and classification scores.
        class_specific_filter: Whether to filter per class, or take best scoring class and filter.
        nms: Flag to enable/disable non maximum suppression.
        score_threshold: Threshold used to prefilter the boxes with.
        max_detections: Maximum number of detections to keep.
        nms_threshold: Threshold for the IoU value to determine when a box should be suppressed.

    Returns
        A list of [boxes, scores, labels, other[0], other[1], ...].
        boxes (max_detections, 4) and contains the (x1, y1, x2, y2) of non-suppressed boxes.
        scores (max_detections,) and contains the scores of the predicted class.
        labels (max_detections,) and contains the predicted label.
        other[i] is shaped (max_detections, ...) and contains the filtered other[i] data.
        In case there are less than max_detections detections, the tensors are padded with -1's.
    """

    def _filter_detections(scores_: tf.Tensor, labels_: tf.Tensor) -> tf.Tensor:
        # threshold based on score
        # (num_score_keeps, 1)
        indices_ = tf.where(keras.backend.greater(scores_, score_threshold))

        if nms:
            # (num_score_keeps, 4)
            filtered_boxes = tf.gather_nd(boxes, indices_)
            # In [4]: scores = np.array([0.1, 0.5, 0.4, 0.2, 0.7, 0.2])
            # In [5]: tf.greater(scores, 0.4)
            # Out[5]: <tf.Tensor: id=2, shape=(6,), dtype=bool,
            # numpy=array([False,  True, False, False,  True, False])>
            # In [6]: tf.where(tf.greater(scores, 0.4))
            # Out[6]:
            # <tf.Tensor: id=7, shape=(2, 1), dtype=int64, numpy=
            # array([[1],
            #        [4]])>
            #
            # In [7]: tf.gather(scores, tf.where(tf.greater(scores, 0.4)))
            # Out[7]:
            # <tf.Tensor: id=15, shape=(2, 1), dtype=float64, numpy=
            # array([[0.5],
            #        [0.7]])>
            filtered_scores = keras.backend.gather(scores_, indices_)[:, 0]

            # perform NMS
            # filtered_boxes = tf.concat([filtered_boxes[..., 1:2], filtered_boxes[..., 0:1],
            #                             filtered_boxes[..., 3:4], filtered_boxes[..., 2:3]],
            # axis=-1)
            nms_indices = tf.image.non_max_suppression(filtered_boxes, filtered_scores,
                                                       max_output_size=max_detections,
                                                       iou_threshold=nms_threshold)

            # filter indices based on NMS
            # (num_score_nms_keeps, 1)
            indices_ = keras.backend.gather(indices_, nms_indices)

        # add indices to list of all indices
        # (num_score_nms_keeps, )
        labels_ = tf.gather_nd(labels_, indices_)
        # (num_score_nms_keeps, 2)
        indices_ = keras.backend.stack([indices_[:, 0], labels_], axis=1)

        return indices_

    if class_specific_filter:
        all_indices = []
        # perform per class filtering
        for category in range(int(classification.shape[1])):
            scores = classification[:, category]
            labels = category * tf.ones((keras.backend.shape(scores)[0],), dtype='int64')
            all_indices.append(_filter_detections(scores, labels))

        # concatenate indices to single tensor
        # (concatenated_num_score_nms_keeps, 2)
        indices = keras.backend.concatenate(all_indices, axis=0)
    else:
        scores = keras.backend.max(classification, axis=1)
        labels = keras.backend.argmax(classification, axis=1)
        indices = _filter_detections(scores, labels)

    # select top k
    scores = tf.gather_nd(classification, indices)
    labels = indices[:, 1]
    scores, top_indices = tf.nn.top_k(scores, k=keras.backend.minimum(
        max_detections, keras.backend.shape(scores)[0]))

    # filter input using the final set of indices
    indices = keras.backend.gather(indices[:, 0], top_indices)
    boxes = keras.backend.gather(boxes, indices)
    labels = keras.backend.gather(labels, top_indices)

    # zero pad the outputs
    pad_size = keras.backend.maximum(0, max_detections - keras.backend.shape(scores)[0])
    boxes = tf.pad(boxes, [[0, pad_size], [0, 0]], constant_values=-1)
    scores = tf.pad(scores, [[0, pad_size]], constant_values=-1)
    labels = tf.pad(labels, [[0, pad_size]], constant_values=-1)
    labels = keras.backend.cast(labels, 'int32')

    # set shapes, since we know what they are
    boxes.set_shape([max_detections, 4])
    scores.set_shape([max_detections])
    labels.set_shape([max_detections])

    if detect_quadrangle:
        alphas = keras.backend.gather(alphas, indices)
        ratios = keras.backend.gather(ratios, indices)
        alphas = tf.pad(alphas, [[0, pad_size], [0, 0]], constant_values=-1)
        ratios = tf.pad(ratios, [[0, pad_size]], constant_values=-1)
        alphas.set_shape([max_detections, 4])
        ratios.set_shape([max_detections])
        return [boxes, scores, alphas, ratios, labels]

    return [boxes, scores, labels]


class FilterDetections(keras.layers.Layer):
    """
    Keras layer for filtering detections using score threshold and NMS.
    """

    def __init__(  # pylint: disable=too-many-arguments
            self,
            nms: bool = True,
            class_specific_filter: bool = True,
            nms_threshold: float = 0.5,
            score_threshold: float = 0.01,
            max_detections: int = 100,
            parallel_iterations: int = 32,
            detect_quadrangle: bool = False,
            **kwargs: Any) -> None:
        """
        Filters detections using score threshold, NMS and selecting the top-k detections.

        Args
            nms: Flag to enable/disable NMS.
            class_specific_filter: To filter per class, or take the best scoring class and filter.
            nms_threshold: Threshold for the IoU value to determine when a box should be suppressed.
            score_threshold: Threshold used to prefilter the boxes with.
            max_detections: Maximum number of detections to keep.
            parallel_iterations: Number of batch items to process in parallel.
        """
        self.nms = nms
        self.class_specific_filter = class_specific_filter
        self.nms_threshold = nms_threshold
        self.score_threshold = score_threshold
        self.max_detections = max_detections
        self.parallel_iterations = parallel_iterations
        self.detect_quadrangle = detect_quadrangle
        super().__init__(**kwargs)

    def call(self, inputs: List[tf.Tensor], **kwargs: Dict[str, Any]) -> tf.Tensor:
        """
        Constructs the NMS graph.

        Args
            inputs : List of [boxes, classification, other[0], other[1], ...] tensors.
        """
        boxes = inputs[0]
        classification = inputs[1]
        if self.detect_quadrangle:
            alphas = inputs[2]
            ratios = inputs[3]

        # wrap nms with our parameters
        def _filter_detections(args: List[tf.Tensor]) -> List[tf.Tensor]:
            boxes_ = args[0]
            classification_ = args[1]
            alphas_ = args[2] if self.detect_quadrangle else None
            ratios_ = args[3] if self.detect_quadrangle else None

            return filter_detections(
                boxes_,
                classification_,
                alphas_,
                ratios_,
                nms=self.nms,
                class_specific_filter=self.class_specific_filter,
                score_threshold=self.score_threshold,
                max_detections=self.max_detections,
                nms_threshold=self.nms_threshold,
                detect_quadrangle=self.detect_quadrangle,
            )

        # call filter_detections on each batch item
        if self.detect_quadrangle:
            outputs = tf.map_fn(
                _filter_detections,
                elems=[boxes, classification, alphas, ratios],
                dtype=['float32', 'float32', 'float32', 'float32', 'int32'],
                parallel_iterations=self.parallel_iterations
            )
        else:
            outputs = tf.map_fn(
                _filter_detections,
                elems=[boxes, classification],
                dtype=['float32', 'float32', 'int32'],
                parallel_iterations=self.parallel_iterations
            )

        return outputs

    def compute_output_shape(self, input_shape: List[tf.TensorShape]) -> tf.TensorShape:
        """
        Computes the output shapes given the input shapes.

        Args
            input_shape : List of input shapes [boxes, classification].

        Returns
            List of tuples representing the output shapes:
            [filtered_boxes.shape, filtered_scores.shape, filtered_labels.shape,
            filtered_other[0].shape, filtered_other[1].shape, ...]
        """
        if self.detect_quadrangle:
            return [
                (input_shape[0][0], self.max_detections, 4),
                (input_shape[1][0], self.max_detections),
                (input_shape[1][0], self.max_detections, 4),
                (input_shape[1][0], self.max_detections),
                (input_shape[1][0], self.max_detections),
            ]

        return [
            (input_shape[0][0], self.max_detections, 4),
            (input_shape[1][0], self.max_detections),
            (input_shape[1][0], self.max_detections),
        ]

    def compute_mask(self, inputs: List[tf.Tensor], mask: tf.Tensor = None) -> tf.Tensor:
        """
        This is required in Keras when there is more than 1 output.
        """
        return (len(inputs) + 1) * [None]

    def get_config(self) -> Dict[str, Any]:
        """
        Gets the configuration of this layer.

        Returns
            Dictionary containing the parameters of this layer.
        """
        config = super().get_config()
        config.update({
            'nms': self.nms,
            'class_specific_filter': self.class_specific_filter,
            'nms_threshold': self.nms_threshold,
            'score_threshold': self.score_threshold,
            'max_detections': self.max_detections,
            'parallel_iterations': self.parallel_iterations,
        })

        return config



"""
Modifications copyright 2021 AI Singapore

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

     https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Original copyright (c) 2019 Zihao Zhang

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import tensorflow as tf


def transform_images(x_image: tf.Tensor, size: int) -> tf.Tensor:
    """transform image to size x size

    Input:
        - x_image: input image matrix
        - size: integer size of the image

    Output:
        - x_image: transformed image matrix
    """
    x_image = tf.image.resize(x_image, (size, size))
    x_image = x_image / 255
    return x_image



"""
Copyright 2021 AI Singapore

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

     https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from typing import List


class QuickFind:
    """
    The general form of the quick-find algorithm that is used to connect nodes.
    How it works: https://www.cs.princeton.edu/~rs/AlgsDS07/01UnionFind.pdf
    Note that although there are faster algorithms such as union-find, we are using
    quick-find here because we need the returning array to have the final group numbers,
    and not have to continue to chase the roots of the nodes subsequently.
    """

    def __init__(self, arr_size: int) -> None:
        self.group_alloc = []
        self.arr_size = arr_size
        for i in range(arr_size):
            self.group_alloc.append(i)

    def get_group_alloc(self) -> List[int]:
        """Getter function for self.group_alloc

        Returns:
            (list): a list containing group numbers allocated to each node.
        """
        return self.group_alloc

    def union(self, node_1_idx: int, node_2_idx: int) -> None:
        """Unites two nodes and other nodes that have been connected to them.

        Args:
            node_1_idx (int): index of the first node
            node_2_idx (int): index of the second node
        """

        node_1_group = self.group_alloc[node_1_idx]
        node_2_group = self.group_alloc[node_2_idx]
        for i in range(self.arr_size):
            if self.group_alloc[i] == node_1_group:
                self.group_alloc[i] = node_2_group

    def connected(self, node_1_idx: int, node_2_idx: int) -> bool:
        """Checks if two nodes are connected.

        Args:
            node_1_idx (int): index of the first node
            node_2_idx (int): index of the second node

        Returns:
            (bool): True if 2 nodes are connected
        """
        return self.group_alloc[node_1_idx] == self.group_alloc[node_2_idx]



"""
Copyright 2021 AI Singapore

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

     https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
from typing import List, Tuple, Any, Iterable, Union
import numpy as np
import cv2
from cv2 import FONT_HERSHEY_SIMPLEX, LINE_AA

POSE_BBOX_COLOR = (255, 255, 0)
BLACK_COLOR = (0, 0, 0)
PINK_COLOR = (255, 0, 255)
ACTIVITY_COLOR = (100, 0, 255)
OBJ_MASK_COLOR = (0, 100, 255)
KEYPOINT_DOT_COLOR = (0, 255, 0)
COUNTING_TEXT_COLOR = (0, 0, 255)
COLOUR_SET_LENGTH = 10
COLOUR_SET = [(255, 0, 0),
              (0, 255, 0),
              (0, 0, 255),
              (255, 255, 0),
              (0, 255, 255),
              (255, 0, 255),
              (255, 165, 0),
              (128, 0, 128),
              (255, 20, 147),
              (255, 250, 250)]
FONT_SCALE = 0.5
FONT_THICKNESS = 1
SKELETON_SHORT_NAMES = (
    "N", "LEY", "REY", "LEA", "REA", "LSH",
    "RSH", "LEL", "REL", "LWR", "RWR",
    "LHI", "RHI", "LKN", "RKN", "LAN", "RAN")

SKELETON = [[16, 14], [14, 12], [17, 15], [15, 13], [12, 13],
            [6, 12], [7, 13], [6, 7], [6, 8], [7, 9],
            [8, 10], [9, 11], [2, 3], [1, 2], [1, 3], [2, 4],
            [3, 5], [4, 6], [5, 7]]


def draw_human_poses(image: np.array,
                     keypoints: np.ndarray,
                     keypoint_scores: np.ndarray,
                     keypoint_conns: np.ndarray,
                     keypoint_dot_color: Tuple[int, int, int],
                     keypoint_dot_radius: int,
                     keypoint_connect_color: Tuple[int, int, int],
                     keypoint_text_color: Tuple[int, int, int]) -> None:
    # pylint: disable=too-many-arguments
    """Draw poses onto an image frame.

    Args:
        image (np.array): image of current frame
        keypoints (List[Any]): list of keypoint coordinates
        keypoints_scores (List[Any]): list of keypoint scores
        keypoints_conns (List[Any]): list of keypoint connections
        keypoint_dot_color (Tuple[int, int, int]): color of keypoint
        keypoint_dot_radius (int): radius of keypoint
        keypoint_connect_color (Tuple[int, int, int]): color of joint
        keypoint_text_color (Tuple[int, int, int]): color of keypoint names
    """
    image_size = _get_image_size(image)
    num_persons = keypoints.shape[0]
    if num_persons > 0:
        for i in range(num_persons):
            _draw_connections(image, keypoint_conns[i],
                              image_size, keypoint_connect_color)
            _draw_keypoints(image, keypoints[i],
                            keypoint_scores[i], image_size,
                            keypoint_dot_color, keypoint_dot_radius, keypoint_text_color)


def _get_image_size(frame: np.array) -> Tuple[int, int]:
    """ Obtain image size of input frame """
    image_size = (frame.shape[1], frame.shape[0])  # width, height
    return image_size


def _draw_connections(frame: np.array,
                      connections: Union[None, Iterable[Any]],
                      image_size: Tuple[int, int],
                      connection_color: Tuple[int, int, int]) -> None:
    """ Draw connections between detected keypoints """
    if connections is not None:
        for connection in connections:
            pt1, pt2 = _project_points_onto_original_image(connection, image_size)
            cv2.line(frame, (pt1[0], pt1[1]), (pt2[0], pt2[1]), connection_color)


def _draw_keypoints(frame: np.ndarray,
                    keypoints: np.ndarray,
                    scores: np.ndarray,
                    image_size: Tuple[int, int],
                    keypoint_dot_color: Tuple[int, int, int],
                    keypoint_dot_radius: int,
                    keypoint_text_color: Tuple[int, int, int]) -> None:
    # pylint: disable=too-many-arguments
    """ Draw detected keypoints """
    img_keypoints = _project_points_onto_original_image(
        keypoints, image_size)

    for idx, keypoint in enumerate(img_keypoints):
        _draw_one_keypoint_dot(frame, keypoint, keypoint_dot_color, keypoint_dot_radius)
        if scores is not None:
            _draw_one_keypoint_text(frame, idx, keypoint, keypoint_text_color)


def _draw_one_keypoint_dot(frame: np.ndarray,
                           keypoint: np.ndarray,
                           keypoint_dot_color: Tuple[int, int, int],
                           keypoint_dot_radius: int) -> None:
    """ Draw single keypoint """
    cv2.circle(frame, (keypoint[0], keypoint[1]), keypoint_dot_radius, keypoint_dot_color, -1)


def _draw_one_keypoint_text(frame: np.ndarray,
                            idx: int,
                            keypoint: np.ndarray,
                            keypoint_text_color: Tuple[int, int, int]) -> None:
    """ Draw name above keypoint """
    position = (keypoint[0], keypoint[1])
    text = str(SKELETON_SHORT_NAMES[idx])

    cv2.putText(frame, text, position, cv2.FONT_HERSHEY_SIMPLEX,
                0.4, keypoint_text_color, 1, cv2.LINE_AA)


def _project_points_onto_original_image(points: np.ndarray,
                                        image_size: Tuple[int, int]) -> np.ndarray:
    """ Project points from relative value (0, 1) to absolute values in original
    image. Note that coordinate (0, 0) starts from image top-left. """
    if len(points) == 0:
        return []

    points = points.reshape((-1, 2))

    projected_points = np.array(points, dtype=np.float32)

    width, height = image_size[0], image_size[1]
    projected_points[:, 0] *= width
    projected_points[:, 1] *= height

    return projected_points


def draw_bboxes(frame: np.array,
                bboxes: List[List[float]],
                color: Tuple[int, int, int],
                thickness: int) -> None:
    """Draw bboxes onto an image frame.

    Args:
        frame (np.array): image of current frame
        bboxes (List[List[float]]): bounding box coordinates
        color (Tuple[int, int, int]): color of bounding box
        thickness (int): thickness of bounding box
    """
    image_size = _get_image_size(frame)
    for bbox in bboxes:
        _draw_bbox(frame, bbox, image_size, color, thickness)


def _draw_bbox(frame: np.array,
               bbox: List[float],
               image_size: Tuple[int, int],
               color: Tuple[int, int, int],
               thickness: int) -> np.array:
    """ Draw a single bounding box """
    top_left, bottom_right = _project_points_onto_original_image(
        bbox, image_size)
    cv2.rectangle(frame, (top_left[0], top_left[1]),
                  (bottom_right[0], bottom_right[1]),
                  color, thickness)


def draw_tags(frame: np.array,
              bboxes: List[List[float]],
              tags: List[str],
              color: Tuple[int, int, int]) -> None:
    """Draw tags above bboxes.

    Args:
        frame (np.array): image of current frame
        bboxes (List[List[float]]): bounding box coordinates
        tags (List[string]): tag associated with bounding box
        color (Tuple[int, int, int]): color of text
    """
    image_size = _get_image_size(frame)
    for idx, bbox in enumerate(bboxes):
        _draw_tag(frame, bbox, tags[idx], image_size, color)


def _draw_tag(frame: np.array,
              bbox: np.array,
              tag: str,
              image_size: Tuple[int, int],
              color: Tuple[int, int, int]) -> None:
    """Draw a tag above a single bounding box.
    """
    top_left, _ = _project_points_onto_original_image(bbox, image_size)
    position = int(top_left[0]), int(top_left[1]-25)
    cv2.putText(frame, tag, position, FONT_HERSHEY_SIMPLEX, 1, color, 2)


def draw_count(frame: np.array, count: int) -> None:
    """draw count of selected object onto frame

    Args:
        frame (np.array): image of current frame
        count (int): total count of selected object
            in current frame
    """
    text = 'COUNT: {0}'.format(count)
    cv2.putText(frame, text, (10, 50), FONT_HERSHEY_SIMPLEX,
                0.75, COUNTING_TEXT_COLOR, 2, LINE_AA)


def draw_pts(frame: np.array, pts: List[Tuple[float]]) -> None:
    """draw pts of selected object onto frame

    Args:
        frame (np.array): image of current frame
        pts (List[Tuple[float]]): bottom midpoints of bboxes
    """
    for point in pts:
        cv2.circle(frame, point, 5, KEYPOINT_DOT_COLOR, -1)


def draw_fps(frame: np.array, current_fps: float) -> None:
    """ Draw FPS onto frame image
<<<<<<< HEAD
    args:
        - frame: array containing the RGB values of the frame image
        - current_fps: value of the calculated FPS
=======

    Args:
        frame (np.array): image of current frame
        current_fps (float): value of the calculated FPS
>>>>>>> 823b177166ef1f40faab8e8611001c73a6b634c7
    """
    text = "FPS: {:.05}".format(current_fps)
    text_location = (25, 25)

    cv2.putText(frame, text, text_location, FONT_HERSHEY_SIMPLEX, FONT_SCALE,
                PINK_COLOR, FONT_THICKNESS, LINE_AA)


def _draw_zone_area(frame: np.array, points: List[Tuple[int]],
                    zone_index: int) -> None:
    total_points = len(points)
    for i in range(total_points):
        if i == total_points-1:
            # for last point, link to first point
            cv2.line(frame, points[i], points[0],
                     COLOUR_SET[zone_index % COLOUR_SET_LENGTH], 3)
        else:
            # for all other points, link to next point in polygon
            cv2.line(frame, points[i], points[i+1],
                     COLOUR_SET[zone_index % COLOUR_SET_LENGTH], 3)


def draw_zones(frame: np.array, zones: List[Any]) -> None:
    """draw the boundaries of the zones used in zoning analytics

    Args:
        frame (np.array): image of current frame
        zones (Zone): zones used in the zoning analytics. possible
        classes are Area and Divider.
    """
    for i, zone_pts in enumerate(zones):
        _draw_zone_area(frame, zone_pts, i)


def draw_zone_count(frame: np.array, zone_count: List[int]) -> None:
    """draw pts of selected object onto frame

    Args:
        frame (np.array): image of current frame
        zone_count (List[float]): object count, likely people, of each zone used
        in the zone analytics
    """
    y_pos = 50
    text = '--ZONE COUNTS--'
    cv2.putText(frame, text, (25, y_pos), FONT_HERSHEY_SIMPLEX, FONT_SCALE,
                COUNTING_TEXT_COLOR, 2, LINE_AA)
    for i, count in enumerate(zone_count):
        y_pos += 25
        text = 'ZONE {0}: {1}'.format(i+1, count)
        cv2.putText(frame, text, (25, y_pos), FONT_HERSHEY_SIMPLEX, FONT_SCALE,
                    COLOUR_SET[i % COLOUR_SET_LENGTH], 2, LINE_AA)



"""
Copyright 2021 AI Singapore

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

     https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
from typing import Tuple

import cv2
import numpy as np

from peekingduck.pipeline.nodes.model.posenetv1.posenet_files.constants import IMAGE_NET_MEAN


def rescale_image(source_img: np.ndarray,
                  input_res: Tuple[int, int],
                  scale_factor: float = 1.0,
                  output_stride: int = 16,
                  model_type: str = 'mobilenet') -> Tuple[np.ndarray, np.ndarray]:
    """Rescale the image by a scale factor while ensuring it has a valid output
    stride

    Args:
        source_img (np.array): image for inference
        input_res (Tuple[int, int]): input height and width of image
        scale_factor (float): ratio to scale image
        output_stride (int): output stride to convert output indices to image coordinates
        model_type (str): specified model type (refer to modelconfig.yml)

    Returns:
        image_processed (np.array): proccessed image array
        scale (np.array): factor to scale height and width
    """
    target_width, target_height = _get_valid_resolution(
        input_res[0] * scale_factor,
        input_res[1] * scale_factor,
        output_stride=output_stride)

    scale = np.array([
        source_img.shape[1] / target_width,
        source_img.shape[0] / target_height
    ])

    image_processed = _rescale_image(source_img, target_width, target_height,
                                     model_type)

    return image_processed, scale


def _get_valid_resolution(width: float,
                          height: float,
                          output_stride: int = 16) -> Tuple[int, int]:
    """Get valid height and width divisible by output stride
    """
    target_width = (int(width) // output_stride) * output_stride + 1
    target_height = (int(height) // output_stride) * output_stride + 1
    return target_width, target_height


def _rescale_image(source_img: np.ndarray,
                   target_width: int,
                   target_height: int,
                   model_type: str) -> np.ndarray:
    """ Apply different preprocessing depending on model type - mobilenet or resnet

    For mobilenet version, RGB values were scaled (2.0 / 255.0) - 1.0
    For resnet version, IMAGE_NET_MEAN were added to the RGB values
    """
    image_processed = cv2.resize(source_img, (target_width, target_height),
                                 interpolation=cv2.INTER_LINEAR)
    image_processed = cv2.cvtColor(image_processed,
                                   cv2.COLOR_BGR2RGB).astype(np.float32)

    if model_type == 'resnet':
        image_processed += IMAGE_NET_MEAN
    else:
        image_processed = image_processed * (2.0 / 255.0) - 1.0

    image_processed = image_processed.reshape(1,
                                              target_height,
                                              target_width,
                                              3)
    return image_processed



"""
Copyright 2021 AI Singapore

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

     https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Code of this file is mostly forked from
[@xuannianz](https://github.com/xuannianz))
"""
from typing import Any, Dict, Callable
import functools
import tensorflow.keras as tfkeras
from peekingduck.pipeline.nodes.model.efficientdet_d04.efficientdet_files \
    import efficientnet as model


def init_tfkeras_custom_objects() -> None:
    """Helper function to initialize custom keras objects
    """
    custom_objects = {
        'swish': inject_tfkeras_modules(model.get_swish)(),
        'FixedDropout': inject_tfkeras_modules(model.get_dropout)()
    }

    tfkeras.utils.get_custom_objects().update(custom_objects)


def inject_tfkeras_modules(func: Callable) -> Callable:
    """Helper function to wrap input function
    """
    @functools.wraps(func)
    def wrapper(*args: Dict[str, Any], **kwargs: Dict[str, Any]) -> Callable:
        kwargs['backend'] = tfkeras.backend
        kwargs['layers'] = tfkeras.layers
        kwargs['models'] = tfkeras.models
        kwargs['utils'] = tfkeras.utils
        return func(*args, **kwargs)

    return wrapper



"""
Copyright 2021 AI Singapore

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
from typing import List, Tuple
from shapely.geometry.polygon import Polygon, Point


class Zone:
    """This class uses polygon area to create a zone for counting.
    """

    def __init__(self, coord_list: List[List[float]]) -> None:
        # Each zone is a polygon created by a list of x, y coordinates
        self.polygon_points = [tuple(x) for x in coord_list]
        self.polygon = Polygon(self.polygon_points)

    def point_within_zone(self, x_coord: float, y_coord: float) -> bool:
        """Function used to check whether the bottom middle point of the bounding box
        is within the stipulated zone created by the divider.

        Args:
            x (float): middle x position of the bounding box
            y (float): lowest y position of the bounding box

        Returns:
            boolean: whether the point given is within the zone.
        """
        return self._is_inside(x_coord, y_coord)

    def get_all_points_of_area(self) -> List[Tuple[float, ...]]:
        """Function used to Get all (x, y) tuple points that form the area of the zone.

        Args:
            None

        Returns:
            list: returns a list of (x, y) points that form the zone area.
        """
        return self.polygon_points

    def _is_inside(self, x_coord: float, y_coord: float) -> bool:
        point = Point((x_coord, y_coord))
        return self.polygon.buffer(1).contains(point)
